# pylint: disable=C0103, C0114, R0913
from __future__ import annotations
from typing import Any
from pathlib import Path
import json
import urllib.request
import warnings

from smufolib import config
from smufolib.utils import error, normalizers

JsonDict = dict[str, Any]

CONFIG = config.load()


class Request:
    """Send HTTP or filesystem request.

    A fallback path (e.g., a filesystem path to the same file), may be
    specified in case of connection failure when the primary path is a
    URL.

    An optional warning may be raised in the event of a fallback.

    :param path: Primary URL or filepath.
    :param fallback: Fallback filepath to use if `path`
        raises :class:`urllib.error.URLError`.
    :param encoding: File text encoding. See :func:`open` for details.
        Defaults to :ref:`[request]` `encoding` configuration.
    :param warn: Warn if URLError is raised before fallback request.
        Defaults to :ref:`[request]` `warn` configuration.
    :param mode: File usage specification used with :attr:`raw`.
        See :func:`open` for details. Defaults to 'r' (read).

    """

    def __init__(
        self,
        path: Path | str | None = None,
        fallback: Path | str | None = None,
        mode: str = "r",
        encoding: str = CONFIG["request"]["encoding"],
        warn: bool = CONFIG["request"]["warn"],
    ) -> None:
        self._path = path
        self._fallback = fallback
        self._mode = mode
        self._encoding = encoding
        self._warn = warn

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} '{self.path}' "
            f"('{self.fallback}') at {id(self)}>"
        )

    def json(self) -> JsonDict | None:
        """Parse request as JSON.

        :raises json.JSONDecodeError: If the response cannot be parsed
            as JSON.
        :raises TypeError: If the raw data is None.

        """
        if self.raw is None:
            return None
        return json.loads(self.raw)

    @property
    def raw(self) -> str | bytes | None:
        """Make a request and return raw file contents.

        :raises ValueError: If both path and fallback are :obj:`None` or
            if the path or fallback file cannot be opened or read.
        :raises urllib.error.URLError: If there is an error with the URL
             request and no fallback is provided.
        :raises FileNotFoundError: If the specified file or fallback
            file cannot be found.

        """
        if self.path is None and self.fallback is None:
            return None

        try:
            if self.path is not None:
                return self._readFromURL()
            return self._readFromFallback()
        except ValueError:
            return self._readFromPath()

    def _readFromURL(self) -> bytes:
        # Read data from URL.
        if self.path is None:
            raise TypeError(
                error.generateTypeError(self.path, (str, Path, Request), "Request.path")
            )
        try:
            with urllib.request.urlopen(self.path) as raw:
                return raw.read()
        except urllib.error.URLError as exc:
            return self._handleURLError(exc)

    def _readFromFallback(self) -> bytes:
        # Read data from fallback file.
        if self.fallback is None:
            raise TypeError(
                error.generateTypeError(
                    self.fallback, (str, Path, Request), "Request.fallback"
                )
            )

        with open(self.fallback, self.mode, encoding=self.encoding) as raw:
            return raw.read()

    def _readFromPath(self) -> bytes:
        # Read data from path.
        if self.path is None:
            raise TypeError(
                error.generateTypeError(self.path, (str, Path, Request), "Request.path")
            )
        with open(self.path, self.mode, encoding=self.encoding) as raw:
            return raw.read()

    def _handleURLError(self, exc: urllib.error.URLError) -> bytes:
        # Handle URL error during online request.
        if not self.fallback:
            raise urllib.error.URLError(
                error.generateErrorMessage("urlError", url=self.path)
            ) from exc

        if self._warn:
            warnings.warn(
                error.generateErrorMessage("urlError", url=self.path), error.URLWarning
            )

        with open(self.fallback, self.mode, encoding=self.encoding) as fallback_file:
            return fallback_file.read()

    @property
    def path(self) -> str | None:
        """Primary URL or filepath as string.

        :raises TypeError: If the path cannot be normalized.

        """
        return normalizers.normalizeRequestPath(self._path, "path")

    @property
    def fallback(self) -> str | None:
        """Fallback URL or filepath as string.

        :raises TypeError: If the fallback path cannot be normalized.

        """
        return normalizers.normalizeRequestPath(self._fallback, "fallback")

    @property
    def mode(self) -> str:
        """File usage specification."""
        return self._mode

    @property
    def encoding(self) -> str:
        """File text encoding."""
        return self._encoding


def writeJson(
    filepath: Path | str,
    source: JsonDict,
    encoding: str = CONFIG["request"]["encoding"],
) -> None:
    """Writes JSON data to filepath.

    :param filepath: Path to target file.
    :param source: JSON data source.
    :param encoding: File text encoding. See :func:`open` for details.
        Defaults to :ref:`[request]` `encoding` configuration.
    :raises TypeError: If `filepath` is not the expected type.
    :raises ValueError: If there is an error serializing the JSON data
        or if `filepath` does not have a ``.json`` exctension.
    :raises FileNotFoundError: If the specified `filepath` cannot be found.
    :raises OSError: If there are any issues opening or writing to the file.
    :raises UnsupportedOperation: If the operation is not supported.

    """
    error.validateType(filepath, (Path, str), "filepath")
    if not str(filepath).endswith(".json"):
        raise ValueError(
            error.generateErrorMessage(
                "missingExtension", objectName="filepath", extension=".json"
            )
        )
    try:
        with open(filepath, "w", encoding=encoding) as outfile:
            json.dump(source, outfile, indent=4, sort_keys=False)
    except (TypeError, ValueError) as e:
        raise ValueError("serializationError") from e



#!/usr/bin/env python3
# coding: utf-8
# pylint: disable=C0301
"""
This script generates a full featured metadata `JSON
<https://www.json.org/json-en.html>`_ file for SMuFL fonts from
metadata sources and user specifications, including the following
metadata sections:

    - `mandatory (fontName and fontVersion)
      <https://w3c.github.io/smufl/latest/specification/font-specific-metadata.html>`_
    - `size <https://w3c.github.io/smufl/latest/specification/font-specific-metadata.html>`_
    - `engravingDefaults <https://w3c.github.io/smufl/latest/specification/engravingdefaults.html>`_
    - `glyphAdvanceWidths <https://w3c.github.io/smufl/latest/specification/glyphadvancewidths.html>`_
    - `glyphsWithAnchors <https://w3c.github.io/smufl/latest/specification/glyphswithanchors.html>`_
    - `glyphsWithAlternates <https://w3c.github.io/smufl/latest/specification/glyphswithalternates.html>`_
    - `glyphBBoxes <https://w3c.github.io/smufl/latest/specification/glyphbboxes.html>`_
    - `ligatures <https://w3c.github.io/smufl/latest/specification/ligatures.html>`_
    - `sets <https://w3c.github.io/smufl/latest/specification/sets.html>`_
    - `optionalGlyphs <https://w3c.github.io/smufl/latest/specification/optionalglyphs.html>`_

Sections are automatically added depending on the font's scope and
assigned attribute values.

.. tip:: As a starting point, identification and engraving defaults
   attributes may be set automatically with the :mod:`~bin.importID`
   and :mod:`~bin.calculateEngravingDefaults` scripts respectively.

Glyphs outside the SMuFL range as well as unencoded or unnamed glyphs,
are automatically skipped.

This script requires SMufoLib to be installed within its executive
environment. It may also be imported as a module and contains the
following public functions:

    - :func:`generateMetadata` - The scripts program function.
    - :func:`main` - Command line entry point.

"""

from __future__ import annotations
from typing import Any
import argparse
import collections
from pathlib import Path

from tqdm import tqdm

from smufolib import Font, config
from smufolib.request import Request, writeJson
from smufolib import cli
from smufolib.utils import stdUtils
from smufolib.utils.scriptUtils import normalizeFont as _normalizeFont
from smufolib.utils.scriptUtils import normalizeJsonDict as _normalizeJsonDict
from smufolib.utils.scriptUtils import normalizeRequest as _normalizeRequest
from smufolib.utils.scriptUtils import normalizeTargetPath as _normalizeTargetPath

# Type aliases
JsonDict = dict[str, Any]
SetsTemplate = dict[str, dict[str, str]]

# Configuration
CONFIG = config.load()

# Parameter defaults
FONT_DATA = Request(
    CONFIG["metadata.paths"]["referenceFont"],
    CONFIG["metadata.fallbacks"]["referenceFont"],
)
VERBOSE = False

# pylint: disable=C0103


def generateMetadata(
    font: Font | Path | str,
    targetPath: str | Path,
    fontData: str | Path | Request = FONT_DATA,
    verbose: bool = VERBOSE,
) -> None:
    """Generate metadata JSON file.

    :param font: Object or path to
        targetPath :class:`~smufolib.objects.font.Font`.
    :param targetPath: Target directory for Metadata JSON file.
    :param fontData: Object call or direct path to reference font
        metadata file. Defaults to :class:`~.Request`
        with :attr:`~.Request.path` and :attr:`~.Request.fallback` set
        to :ref:`[metadata.paths]` and :ref:`
        [metadata.fallbacks]` respective `referenceFont` configurations.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type.
    :raises FileNotFoundError: If `targetPath` does not exist.

    """
    print("Starting...")

    targetPath = _normalizeTargetPath(targetPath)
    font = _normalizeFont(font)
    fontDataJson = _normalizeJsonDict(_normalizeRequest(fontData).json())
    font.smufl.spaces = True

    metadata = _compileMetadata(font, fontDataJson, verbose)

    filename = Path(targetPath).resolve() / f"{font.smufl.name}.json"
    stdUtils.verbosePrint(f"\nWriting metadata to: '{filename}'", verbose)
    writeJson(filename, metadata)

    print("\nDone.")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    generateMetadata(
        args.font, args.targetPath, fontData=args.fontData, verbose=args.verbose
    )


def _compileMetadata(font: Font, fontData: JsonDict, verbose: bool) -> JsonDict:
    # Build SMuFL-specific metadata dictionary.
    # pylint: disable=R0912, R0914
    stdUtils.verbosePrint("\nCompiling font metadata...", verbose)

    metadata: collections.defaultdict[str, JsonDict] = collections.defaultdict(dict)

    keys = ("fontName", "fontVersion", "designSize", "sizeRange", "engravingDefaults")

    for key in keys:
        attribute = key[4:].lower() if "font" in key else key
        value = getattr(font.smufl, attribute)
        if attribute == "engravingDefaults":
            value = value.items()
        elif value is None:
            continue
        metadata[key] = value

    sets, setGlyphs = {}, []
    setsTemplate = _getSetsTemplate(fontData)

    stdUtils.verbosePrint("\nCompiling metadata for glyphs:", verbose)
    for glyph in font if verbose else tqdm(font):
        if glyph is None or not stdUtils.validateAttr(
            glyph, ("name", "unicode", "smufl.isMember")
        ):
            stdUtils.verbosePrint("\tSkipping invalid glyph:", verbose, glyph)
            continue
        stdUtils.verbosePrint(f"\t{glyph}", verbose)

        s = glyph.smufl
        metadata["glyphAdvanceWidths"][s.name] = s.advanceWidth
        metadata["glyphBBoxes"][s.name] = s.bBox

        if s.alternates:
            metadata["glyphsWithAlternates"][s.name] = {"alternates": s.alternates}

        if s.anchors:
            metadata["glyphsWithAnchors"][s.name] = s.anchors

        if s.isLigature:
            metadata["ligatures"][s.name] = {
                "codepoint": s.codepoint,
                "componentGlyphs": s.componentNames,
                "description": s.description,
            }

        if s.isOptional:
            metadata["optionalGlyphs"][s.name] = {
                "codepoint": s.codepoint,
                "description": s.description,
                "classes": s.classes,
            }

        if s.isSet:
            sets[s.suffix] = setsTemplate[s.suffix]
            glyphInfo = {
                "alternateFor": s.base.smufl.name,
                "codepoint": s.codepoint,
                "description": s.description,
                "name": s.name,
            }
            setGlyphs.append(glyphInfo)
            metadata["sets"] = sets
            metadata["sets"][s.suffix]["glyphs"] = setGlyphs

    return metadata


def _getSetsTemplate(fontData: JsonDict) -> SetsTemplate:
    # Extract sets template from bravura_metadata.json.
    sets: dict[str, dict[str, str]] = {}

    for key, values in fontData["sets"].items():
        sets[key] = {}
        for subKey, value in values.items():
            if subKey not in ("description", "type"):
                continue
            sets[key][subKey] = value
    return sets


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        "font",
        "targetPath",
        description=stdUtils.getSummary(generateMetadata.__doc__),
        fontData=FONT_DATA,
        verbose=VERBOSE,
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



# pylint: disable=C0114, C0103, W0212, W0221
from __future__ import annotations
from typing import TYPE_CHECKING

from fontParts.base.base import BaseObject
from smufolib.utils import error, normalizers

if TYPE_CHECKING:  # pragma: no cover
    from smufolib.objects.smufl import Smufl
    from smufolib.objects.glyph import Glyph
    from smufolib.objects.font import Font
    from smufolib.objects.layer import Layer

# Type aliases
EngravingDefaultsValue = int | float | tuple[str, ...] | None
EngravingDefaultsDict = dict[str, EngravingDefaultsValue]

#: Names of engraving defaults as specified in the SMuFL standard.
ENGRAVING_DEFAULTS_KEYS: set = {
    "arrowShaftThickness",
    "barlineSeparation",
    "beamSpacing",
    "beamThickness",
    "bracketThickness",
    "dashedBarlineDashLength",
    "dashedBarlineGapLength",
    "dashedBarlineThickness",
    "hairpinThickness",
    "hBarThickness",
    "legerLineExtension",
    "legerLineThickness",
    "lyricLineThickness",
    "octaveLineThickness",
    "pedalLineThickness",
    "repeatBarlineDotSeparation",
    "repeatEndingLineThickness",
    "slurEndpointThickness",
    "slurMidpointThickness",
    "staffLineThickness",
    "stemThickness",
    "subBracketThickness",
    "textFontFamily",
    "textEnclosureThickness",
    "thickBarlineThickness",
    "thinBarlineThickness",
    "thinThickBarlineSeparation",
    "tieEndpointThickness",
    "tieMidpointThickness",
    "tupletBracketThickness",
}


class EngravingDefaults(BaseObject):
    """SMuFL engraving default settings.

    This object contains properties and methods pertained to SMuFL's
    `engravingDefaults
    <https://w3c.github.io/smufl/latest/specification/engravingdefaults.html>`_
    metadata structure, defining recommended defaults for line widths
    etc., according to the specification.

    The :class:`EngravingDefaults` object is essentially a :class:`dict`
    with access to keys through regular class attributes.

    .. tip:: Engraving default values can be calculated from the
       contours of specific glyphs and set automatically with
       the :mod:`~bin.calculateEngravingDefaults` script.

    :param smufl: Parent :class:`~smufolib.objects.smufl.Smufl` object.

    While this object is normally created as part of
    a :class:`~smufolib.objects.font.Font`, an
    orphan :class:`EngravingDefaults` object may be created like
    this::

        >>> d = EngravingDefaults()

    """

    def _init(self, smufl: Smufl | None = None) -> None:
        self._smufl = smufl

    def _reprContents(self) -> list[str]:
        contents = []
        if self.font is not None:
            contents.append("in font")

            contents += self.font._reprContents()  # pylint: disable-next=W0212
        return contents

    def naked(self):
        # BaseObject override for __eq__ and __hash__
        return self  # pragma: no cover

    # -------
    # Parents
    # -------

    @property
    def smufl(self) -> Smufl | None:
        """Parent :class:`~smufolib.objects.smufl.Smufl`."""
        return self._smufl

    @smufl.setter
    def smufl(self, value: Smufl) -> None:
        if self._smufl is not None and self._smufl != value:
            raise AssertionError(
                "Smufl for EngravingDefaults object is "
                "already set and is not same as value."
            )
        self._smufl = normalizers.normalizeSmufl(value)

    @property
    def font(self) -> Font | None:
        """Parent :class:`~smufolib.objects.font.Font` object."""
        if self._smufl is None:
            return None
        return self._smufl.font

    @property
    def glyph(self) -> Glyph | None:
        """Parent :class:`~smufolib.objects.glyph.Glyph` object."""
        if self._smufl is None:
            return None
        return self._smufl.glyph

    @property
    def layer(self) -> Layer | None:
        """Parent :class:`~smufolib.objects.layer.Layer` object."""
        if self._smufl is None:
            return None
        return self._smufl.layer

    # ------------------
    # Dictionary methods
    # ------------------

    def clear(self) -> None:
        """Clear all engraving default settings.

        Example::

            >>> d.items()
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}
            >>> d.clear()
            >>> d.items()
            {'arrowShaftThickness': None, 'barlineSeparation': None, ...}

        """
        if self.font is not None:
            self.font.lib.naked().pop("com.smufolib.engravingDefaults", None)

    def items(self) -> EngravingDefaultsDict:
        """Return dict of all available settings and their values.

        Example::

            >>> d = f.smufl.engravingDefaults
            >>> d.items()
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}

        """
        return dict(zip(self.keys(), self.values()))

    def keys(self) -> list[str]:
        """Return sorted list of all available settings names.

        Example::

            >>> d = f.smufl.engravingDefaults
            >>> d.keys()
            ['arrowShaftThickness', 'barlineSeparation', ...]

        """
        return sorted(ENGRAVING_DEFAULTS_KEYS)

    def update(
        self,
        other: EngravingDefaults | EngravingDefaultsDict | None = None,
        **kwargs: EngravingDefaultsValue,
    ) -> None:
        r"""Update settings attributes with other object or values.

        :param other: Other :class:`EngravingDefaults` or :class:`dict`
            of attribute names mapped to values. Defaults to `None`.
        :param \**kwargs: Attribute names and values to update as
         keyword arguments.

        An object may be updated in a few different ways::

            >>> d1 = f.smufl.engravingDefaults
            >>> d1.items()
            {'arrowShaftThickness': None, 'barlineSeparation': None, ...}

        Update with :class:`dict` object.::

            >>> d2 = {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}
            >>> d1.update(d2)
            >>> d1.items()
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}

        Update with other :class:`EngravingDefaults` object.::

            >>> d2 = f2.smufl.engravingDefaults
            >>> d2.items()
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}
            >>> d1.update(d2)
            >>> d1.items()
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}

        Update with keyword arguments::

            >>> d1.update(arrowShaftThickness=46, barlineSeparation=72)
            >>> d1.items()
            {'arrowShaftThickness': 46, 'barlineSeparation': 72, ...}

        """
        other = self._normalizeOthers(other, kwargs)
        if not other or self.font is None:
            return

        if "com.smufolib.engravingDefaults" not in self.font.lib:
            self.font.lib["com.smufolib.engravingDefaults"] = {}

        for key, value in other.items():  # type: ignore[misc]
            if key not in self.keys():
                raise AttributeError(
                    error.generateErrorMessage(
                        "attributeError", objectName=type(self).__name__, attribute=key
                    )
                )

            normalizedValue = normalizers.normalizeEngravingDefaultsAttr(key, value)
            if value is None:
                continue
            if (
                self._smufl is not None
                and isinstance(normalizedValue, (int, float))
                and self.spaces
            ):
                normalizedValue = self._smufl.toUnits(normalizedValue)
            self.font.lib["com.smufolib.engravingDefaults"][key] = normalizedValue

    def _normalizeOthers(
        self,
        other: EngravingDefaults | EngravingDefaultsDict | None,
        kwargs: EngravingDefaultsDict,
    ) -> EngravingDefaultsDict | None:
        # Normalize update(other)
        if isinstance(other, EngravingDefaults):
            other = other.items()
        if other is None:
            other = {}
            if not kwargs:
                return None

        error.validateType(
            other,
            (EngravingDefaults, dict, type(None)),
            f"{self.__class__.__name__}.update()",
        )

        other |= kwargs
        return other

    def values(self) -> list[EngravingDefaultsValue]:
        """Return list of all available settings values.

        Order corresponds to :meth:`keys`.

        Example::

            >>> d = f.smufl.engravingDefaults
            >>> d.values()
            [46, 72, 62, 125, 125, 125, 57, 36, 200, 39, 80, 40, ...]

        """
        return [getattr(self, k) for k in self.keys()]

    # -------------------
    # Settings Properties
    # -------------------

    @property
    def arrowShaftThickness(self) -> int | float | None:
        """Thickness of the line used for the shaft of an arrow."""
        return self._getValue("arrowShaftThickness")  # type: ignore

    @arrowShaftThickness.setter
    def arrowShaftThickness(self, value: int | float | None) -> None:
        self._setValue("arrowShaftThickness", value)

    @property
    def barlineSeparation(self) -> int | float | None:
        """Distance between multiple thin barlines."""
        return self._getValue("barlineSeparation")  # type: ignore

    @barlineSeparation.setter
    def barlineSeparation(self, value: int | float | None) -> None:
        self._setValue("barlineSeparation", value)

    @property
    def beamSpacing(self) -> int | float | None:
        """Distance between beams."""
        return self._getValue("beamSpacing")  # type: ignore

    @beamSpacing.setter
    def beamSpacing(self, value: int | float | None) -> None:
        self._setValue("beamSpacing", value)

    @property
    def beamThickness(self) -> int | float | None:
        """Thickness of a beam."""
        return self._getValue("beamThickness")  # type: ignore

    @beamThickness.setter
    def beamThickness(self, value: int | float | None) -> None:
        self._setValue("beamThickness", value)

    @property
    def bracketThickness(self) -> int | float | None:
        """Thickness of the vertical line of a grouping bracket."""
        return self._getValue("bracketThickness")  # type: ignore

    @bracketThickness.setter
    def bracketThickness(self, value: int | float | None) -> None:
        self._setValue("bracketThickness", value)

    @property
    def dashedBarlineDashLength(self) -> int | float | None:
        """Length of the dashes to be used in a dashed barline."""
        return self._getValue("dashedBarlineDashLength")  # type: ignore

    @dashedBarlineDashLength.setter
    def dashedBarlineDashLength(self, value: int | float | None) -> None:
        self._setValue("dashedBarlineDashLength", value)

    @property
    def dashedBarlineGapLength(self) -> int | float | None:
        """Length of the gap between dashes in a dashed barline."""
        return self._getValue("dashedBarlineGapLength")  # type: ignore

    @dashedBarlineGapLength.setter
    def dashedBarlineGapLength(self, value: int | float | None) -> None:
        self._setValue("dashedBarlineGapLength", value)

    @property
    def dashedBarlineThickness(self) -> int | float | None:
        """Thickness of a dashed barline."""
        return self._getValue("dashedBarlineThickness")  # type: ignore

    @dashedBarlineThickness.setter
    def dashedBarlineThickness(self, value: int | float | None) -> None:
        self._setValue("dashedBarlineThickness", value)

    @property
    def hairpinThickness(self) -> int | float | None:
        """Thickness of a crescendo/diminuendo hairpin."""
        return self._getValue("hairpinThickness")  # type: ignore

    @hairpinThickness.setter
    def hairpinThickness(self, value: int | float | None) -> None:
        self._setValue("hairpinThickness", value)

    @property
    def hBarThickness(self) -> int | float | None:
        """Thickness of a crescendo/diminuendo hairpin."""
        return self._getValue("hBarThickness")  # type: ignore

    @hBarThickness.setter
    def hBarThickness(self, value: int | float | None) -> None:
        self._setValue("hBarThickness", value)

    @property
    def legerLineExtension(self) -> int | float | None:
        """Amount of leger line extension from notehead."""
        return self._getValue("legerLineExtension")  # type: ignore

    @legerLineExtension.setter
    def legerLineExtension(self, value: int | float | None) -> None:
        self._setValue("legerLineExtension", value)

    @property
    def legerLineThickness(self) -> int | float | None:
        """Thickness of a leger line."""
        return self._getValue("legerLineThickness")  # type: ignore

    @legerLineThickness.setter
    def legerLineThickness(self, value: int | float | None) -> None:
        self._setValue("legerLineThickness", value)

    @property
    def lyricLineThickness(self) -> int | float | None:
        """Thickness of the lyric extension line."""
        return self._getValue("lyricLineThickness")  # type: ignore

    @lyricLineThickness.setter
    def lyricLineThickness(self, value: int | float | None) -> None:
        self._setValue("lyricLineThickness", value)

    @property
    def octaveLineThickness(self) -> int | float | None:
        """Thickness of the dashed line used for an octave line."""
        return self._getValue("octaveLineThickness")  # type: ignore

    @octaveLineThickness.setter
    def octaveLineThickness(self, value: int | float | None) -> None:
        self._setValue("octaveLineThickness", value)

    @property
    def pedalLineThickness(self) -> int | float | None:
        """Thickness of the line used for piano pedaling."""
        return self._getValue("pedalLineThickness")  # type: ignore

    @pedalLineThickness.setter
    def pedalLineThickness(self, value: int | float | None) -> None:
        self._setValue("pedalLineThickness", value)

    @property
    def repeatBarlineDotSeparation(self) -> int | float | None:
        """Distance between dots and inner line of a repeat barline."""
        return self._getValue("repeatBarlineDotSeparation")  # type: ignore

    @repeatBarlineDotSeparation.setter
    def repeatBarlineDotSeparation(self, value: int | float | None) -> None:
        self._setValue("repeatBarlineDotSeparation", value)

    @property
    def repeatEndingLineThickness(self) -> int | float | None:
        """Thickness of repeat ending brackets."""
        return self._getValue("repeatEndingLineThickness")  # type: ignore

    @repeatEndingLineThickness.setter
    def repeatEndingLineThickness(self, value: int | float | None) -> None:
        self._setValue("repeatEndingLineThickness", value)

    @property
    def slurEndpointThickness(self) -> int | float | None:
        """Thickness of the end of a slur."""
        return self._getValue("slurEndpointThickness")  # type: ignore

    @slurEndpointThickness.setter
    def slurEndpointThickness(self, value: int | float | None) -> None:
        self._setValue("slurEndpointThickness", value)

    @property
    def slurMidpointThickness(self) -> int | float | None:
        """Thickness of the mid-point of a slur."""
        return self._getValue("slurMidpointThickness")  # type: ignore

    @slurMidpointThickness.setter
    def slurMidpointThickness(self, value: int | float | None) -> None:
        self._setValue("slurMidpointThickness", value)

    @property
    def staffLineThickness(self) -> int | float | None:
        """Thickness of each staff line."""
        return self._getValue("staffLineThickness")  # type: ignore

    @staffLineThickness.setter
    def staffLineThickness(self, value: int | float | None) -> None:
        self._setValue("staffLineThickness", value)

    @property
    def stemThickness(self) -> int | float | None:
        """Thickness of a stem."""
        return self._getValue("stemThickness")  # type: ignore

    @stemThickness.setter
    def stemThickness(self, value: int | float | None) -> None:
        self._setValue("stemThickness", value)

    @property
    def subBracketThickness(self) -> int | float | None:
        """Thickness of the vertical line of a grouping sub-bracket."""
        return self._getValue("subBracketThickness")  # type: ignore

    @subBracketThickness.setter
    def subBracketThickness(self, value: int | float | None) -> None:
        self._setValue("subBracketThickness", value)

    @property
    def textFontFamily(self) -> tuple[str, ...] | None:
        """tuple of text font pairings."""
        return self._getValue("textFontFamily")  # type: ignore

    @textFontFamily.setter
    def textFontFamily(self, value: tuple[str, ...]) -> None:
        self._setValue("textFontFamily", value)

    @property
    def textEnclosureThickness(self) -> int | float | None:
        """Thickness of box drawn around text instructions."""
        return self._getValue("textEnclosureThickness")  # type: ignore

    @textEnclosureThickness.setter
    def textEnclosureThickness(self, value: int | float | None) -> None:
        self._setValue("textEnclosureThickness", value)

    @property
    def thickBarlineThickness(self) -> int | float | None:
        """Thickness of a thick barline."""
        return self._getValue("thickBarlineThickness")  # type: ignore

    @thickBarlineThickness.setter
    def thickBarlineThickness(self, value: int | float | None) -> None:
        self._setValue("thickBarlineThickness", value)

    @property
    def thinBarlineThickness(self) -> int | float | None:
        """Thickness of a thin barline."""
        return self._getValue("thinBarlineThickness")  # type: ignore

    @thinBarlineThickness.setter
    def thinBarlineThickness(self, value: int | float | None) -> None:
        self._setValue("thinBarlineThickness", value)

    @property
    def thinThickBarlineSeparation(self) -> int | float | None:
        """Thickness of a thin barline."""
        return self._getValue("thinThickBarlineSeparation")  # type: ignore

    @thinThickBarlineSeparation.setter
    def thinThickBarlineSeparation(self, value: int | float | None) -> None:
        self._setValue("thinThickBarlineSeparation", value)

    @property
    def tieEndpointThickness(self) -> int | float | None:
        """Thickness of the end of a tie."""
        return self._getValue("tieEndpointThickness")  # type: ignore

    @tieEndpointThickness.setter
    def tieEndpointThickness(self, value: int | float | None) -> None:
        self._setValue("tieEndpointThickness", value)

    @property
    def tieMidpointThickness(self) -> int | float | None:
        """Thickness of the mid-point of a tie."""
        return self._getValue("tieMidpointThickness")  # type: ignore

    @tieMidpointThickness.setter
    def tieMidpointThickness(self, value: int | float | None) -> None:
        self._setValue("tieMidpointThickness", value)

    @property
    def tupletBracketThickness(self) -> int | float | None:
        """Thickness of tuplet brackets."""
        return self._getValue("tupletBracketThickness")  # type: ignore

    @tupletBracketThickness.setter
    def tupletBracketThickness(self, value: int | float | None) -> None:
        self._setValue("tupletBracketThickness", value)

    def _getValue(self, name: str) -> EngravingDefaultsValue:
        # Common settings property getter.
        if not self._libDict:
            return None

        value = self._libDict.get(name, None)
        if name == "textFontFamily":
            value = self._libDict.get(name, ())

        elif self.spaces and isinstance(value, (int, float)) and self.font is not None:
            return self.font.smufl.toSpaces(value)
        return value

    def _setValue(self, name: str, value: EngravingDefaultsValue) -> None:
        # Common settings property setter.
        # Keeps dynamic dict of settings in font.lib().

        if self.font is None:
            return

        normalizedValue = normalizers.normalizeEngravingDefaultsAttr(name, value)
        libDict = self._libDict

        if not normalizedValue:
            libDict.pop(name, None)

            if not libDict:
                self.font.lib.pop("com.smufolib.engravingDefaults", None)
            else:
                self._libDict = libDict

        else:
            if self.spaces and isinstance(normalizedValue, (int, float)):
                normalizedValue = self.font.smufl.toUnits(normalizedValue)

            self._libDict[name] = normalizedValue
            self._libDict = libDict

    @property
    def _libDict(self) -> dict[str, EngravingDefaultsValue]:
        # Dynamic dict in font.lib.naked().
        if self.font is None:
            return {}
        lib = self.font.lib.naked()
        return lib.setdefault("com.smufolib.engravingDefaults", {})

    @_libDict.setter
    def _libDict(self, value: EngravingDefaultsDict) -> None:
        if self.font is not None:
            self.font.lib.naked()["com.smufolib.engravingDefaults"] = value

    # -----------------------------
    # Normalization and Measurement
    # -----------------------------

    def round(self) -> None:
        """Round font units to integers.

        Method applies to the following attributes:

        * :attr:`arrowShaftThickness`
        * :attr:`barlineSeparation`
        * :attr:`beamSpacing`
        * :attr:`beamThickness`
        * :attr:`bracketThickness`
        * :attr:`dashedBarlineDashLength`
        * :attr:`dashedBarlineGapLength`
        * :attr:`dashedBarlineThickness`
        * :attr:`hairpinThickness`
        * :attr:`hBarThickness`
        * :attr:`legerLineExtension`
        * :attr:`legerLineThickness`
        * :attr:`lyricLineThickness`
        * :attr:`octaveLineThickness`
        * :attr:`pedalLineThickness`
        * :attr:`repeatBarlineDotSeparation`
        * :attr:`repeatEndingLineThickness`
        * :attr:`slurEndpointThickness`
        * :attr:`slurMidpointThickness`
        * :attr:`staffLineThickness`
        * :attr:`stemThickness`
        * :attr:`subBracketThickness`
        * :attr:`textEnclosureThickness`
        * :attr:`thickBarlineThickness`
        * :attr:`thinBarlineThickness`
        * :attr:`thinThickBarlineSeparation`
        * :attr:`tieEndpointThickness`
        * :attr:`tieMidpointThickness`
        * :attr:`tupletBracketThickness`

        If :attr:`spaces` is :obj:`True`, values are left unchanged.

        Example::

            >>> f.smufl.engravingDefaults.stemThickness
            30.5
            >>> f.smufl.engravingDefaults.round()
            >>> f.smufl.engravingDefaults.stemThickness
            31

        ::

            >>> f.smufl.engravingDefaults.spaces = True
            >>> f.smufl.engravingDefaults.stemThickness
            0.12
            >>> f.smufl.engravingDefaults.round()
            >>> f.smufl.engravingDefaults.stemThickness
            0.12

        """
        if self.spaces:
            return

        for key in self.keys():
            attribute = getattr(self, key)
            if not isinstance(attribute, (int, float)):
                continue
            value = normalizers.normalizeVisualRounding(attribute)
            setattr(self, key, value)

    @property
    def spaces(self) -> bool:
        """Set state of measurement to staff spaces.

        Example::

            >>> f.smufl.engravingDefaults.stemThickness
            30.5
            >>> f.smufl.engravingDefaults.spaces = True
            >>> f.smufl.engravingDefaults.stemThickness
            0.12

        """
        if self._smufl is None:
            return False
        return self._smufl.spaces

    @spaces.setter
    def spaces(self, value: bool) -> None:
        if self._smufl is None:
            return
        self._smufl.spaces = value

    # ------------------------
    # Override from BaseObject
    # ------------------------

    def raiseNotImplementedError(self):
        """This exception needs to be raised frequently by
        the base classes. So, it's here for convenience.
        """
        raise NotImplementedError(  # pragma: no cover
            error.generateErrorMessage(
                "notImplementedError", objectName=self.__class__.__name__
            )
        )



"""Utilities for extracting and representing point data.

This module provides classes and functions to retrieve points from
contours and components within font-related objects and to simplify
the representation of :class:`fontParts.base.BasePoint`. It includes
utilities for handling both direct glyph shapes and composite glyphs,
supporting point types, positions, and contour indices.

"""

from __future__ import annotations
from typing import NamedTuple
from collections.abc import Iterator
import itertools

from fontParts.fontshell import RComponent
from fontParts.fontshell import RContour
from smufolib.objects.glyph import Glyph
from smufolib.objects.layer import Layer
from smufolib.objects.font import Font
from smufolib.utils import error, stdUtils

TYPES = ("line", "curve", "qcurve")

# pylint: disable=C0103


class Point(NamedTuple):
    """Named tuple for point values.

    :param type: The :attr:`~fontParts.base.BasePoint.type` of the
        point.
    :param position: The :class:`~Position` of the point.
    :param contourIndex: The :attr:`~fontParts.base.BaseContour.index`
        of the point's parent contour.

    """

    type: str
    position: Position
    contourIndex: int


class Position(NamedTuple):
    """Named tuple for position values.

    :param x: The horizontal position of the point.
    :param y: The vertical position of the point.

    """

    x: int | float
    y: int | float


def getPoints(
    obj: Glyph | tuple[Glyph, ...] | Layer | Font, types: str | tuple[str, ...] = TYPES
) -> Iterator[Point]:
    """Get points from font-related object.

    This function simply concatinates the results
    of :func:`getContourPoints` and :func:`getCompositePoints` and
    returns the result as an :obj:`itertools.chain` object.

    :param obj: The object from which to get the points. This can be a
        single :class:`.Glyph`, a :class:`tuple` of glyph objects or any
        parent object (:class:`.Layer` or :class:`.Font`).
    :param types: The :attr:`fontParts.base.BasePoint.type` to include.
        Defaults to ``('line', 'curve', 'qcurve')``.
    :raises TypeError: If `obj` is not an accepted type.
    :raises ValueError: If `obj` items is not an accepted type.

    Example::

        >>> glyph = font['uniE071'] # schaefferGClefToFClef (contour and component parts)
        >>> points = pointUtils.getPoints(glyph)
        >>> tuple(points)
        (Point(type='line', position=Position(x=618, y=0), contourIndex=0),
        Point(type='line', position=Position(x=309, y=250), contourIndex=0),
        Point(type='line', position=Position(x=309, y=-249), contourIndex=0),
        Point(type='line', position=Position(x=309, y=-749), contourIndex=1),
        Point(type='line', position=Position(x=309, y=-250), contourIndex=1),
        Point(type='line', position=Position(x=0, y=-500), contourIndex=1),
        Point(type='line', position=Position(x=281, y=-691), contourIndex=2),
        Point(type='line', position=Position(x=45, y=-500), contourIndex=2),
        Point(type='line', position=Position(x=281, y=-309), contourIndex=2))

    """
    error.validateType(obj, (tuple, Glyph, Layer, Font), "obj")
    if isinstance(obj, tuple):
        for item in obj:
            error.validateType(item, Glyph, "obj", items=True)
        return itertools.chain.from_iterable(
            itertools.chain(getContourPoints(g, types), getCompositePoints(g, types))
            for g in obj
        )
    return itertools.chain(getContourPoints(obj, types), getCompositePoints(obj, types))


def getContourPoints(
    obj: RContour | tuple[RContour, ...] | Glyph | Layer | Font,
    types: str | tuple[str, ...] = TYPES,
) -> Iterator[Point]:
    """Get contour points from a font-related object.

    Contour points are those that define the shape of a glyph directly
    through its outline.

    :param obj: The object from which to get the points. This can be a
        single contour object, a :class:`tuple` of contours or any
        parent object (:class:`.Glyph`, :class:`.Layer`
        or :class:`.Font`).
    :param types: The :attr:`fontParts.base.BasePoint.type` to include.
        Defaults to ``('line', 'curve', 'qcurve')``.
    :raises TypeError: If `obj` is not an accepted type.

    Example::

        >>> glyph = font['uniE071'] # schaefferGClefToFClef (contour part)
        >>> points = pointUtils.getContourPoints(glyph)
        >>> tuple(points)
        (Point(type='line', position=Position(x=618, y=0), contourIndex=0),
        Point(type='line', position=Position(x=309, y=250), contourIndex=0),
        Point(type='line', position=Position(x=309, y=-249), contourIndex=0))

    """
    error.validateType(obj, (tuple, RContour, Glyph, Layer, Font), "obj")
    if isinstance(obj, tuple):
        for component in obj:
            error.validateType(component, RContour, "obj", items=True)
    if isinstance(obj, RContour):
        obj = (obj,)

    rawPoints = stdUtils.flatten(obj)

    return (
        Point(p.type, Position(*p.position), p.contour.index)
        for p in rawPoints
        if p.type in types or p.type == types
    )


def getCompositePoints(
    obj: RComponent | tuple[RComponent, ...] | Glyph | Layer | Font,
    types: str | tuple[str, ...] = TYPES,
) -> Iterator[Point]:
    """Get composite points from font-related object.

    Composite points are those that come from glyphs that are used as
    components in building more complex glyphs.

    The component's offset position is taken into account.

    :param obj: The object from which to get the points. This can be a
        single component object, a :class:`tuple` of components or any
        parent object (:class:`.Glyph`, :class:`.Layer`
        or :class:`.Font`).
    :param types: The :attr:`fontParts.base.BasePoint.type` to include.
        Defaults to ``('line', 'curve', 'qcurve')``.
    :raises TypeError: If `obj` is not an accepted type.

    Example::

        >>> glyph = font['uniE071'] # schaefferGClefToFClef (component part)
        >>> points = pointUtils.getCompositePoints(glyph)
        >>> tuple(points)
        (Point(type='line', position=Position(x=309, y=-749), contourIndex=1),
        Point(type='line', position=Position(x=309, y=-250), contourIndex=1),
        Point(type='line', position=Position(x=0, y=-500), contourIndex=1),
        Point(type='line', position=Position(x=281, y=-691), contourIndex=2),
        Point(type='line', position=Position(x=45, y=-500), contourIndex=2),
        Point(type='line', position=Position(x=281, y=-309), contourIndex=2))

    """
    components = _getComponents(obj)

    def pointGenerator():
        for component in components:
            baseGlyph = component.font[component.baseGlyph]
            rawPoints = stdUtils.flatten(baseGlyph.contours)
            offset, index = component.offset, component.index
            for point in rawPoints:
                if point.type in types or point.type == types:
                    newPosition = stdUtils.addTuples(point.position, offset)
                    yield Point(point.type, Position(*newPosition), index)

    return pointGenerator()


def _getComponents(
    obj: RComponent | tuple[RComponent, ...] | Glyph | Layer | Font,
) -> tuple[RComponent, ...] | Iterator[RComponent]:
    # Get components from font-related object.
    if isinstance(obj, tuple):
        for component in obj:
            error.validateType(component, RComponent, "obj", items=True)
        return obj
    if isinstance(obj, RComponent):
        return (obj,)
    if isinstance(obj, Glyph):
        return obj.components
    if isinstance(obj, (Layer, Font)):
        return stdUtils.flatten(g.components for g in obj)
    raise TypeError(
        error.generateTypeError(obj, (RComponent, tuple, Glyph, Layer, Font), "obj")
    )



# pylint: disable=C0114
from fontParts.fontshell.glyph import RGlyph
from smufolib.objects.smufl import Smufl


class Glyph(RGlyph):
    """Environment implementation of :class:`fontParts.base.BaseGlyph`.

    Glyphs are usually accessed through
    a :class:`~smufolib.objects.font.Font` objects inherent glyph
    dictionary. To instantiate the SMuFL glyph named `uniE030`
    (single barline)::

        >>> glyph = font['uniE030']

    """

    def _set_name(self, value):
        # Set the name of the glyph and update :attr:`.Font.lib` if necessary.
        if self.font is not None:
            namesDict = self.font.lib.get("com.smufolib.names", {})
            smuflName = self.smufl.name
            if smuflName and self.naked().name in namesDict.values():
                namesDict[smuflName] = value
        self.naked().name = value

    @property
    def smufl(self) -> Smufl:
        """Glyph instance of :class:`~smufolib.objects.smufl.Smufl`."""
        return Smufl(glyph=self)



#!/usr/bin/env python3
# coding: utf-8
"""
This script deletes SMuFL-specific metadata and glyph anchors from a UFO
font file.

To prevent unwanted data loss, items to delete *must* be specifically
included and may also be specifically excluded. If ``exclude=None``,
``include='*'`` (all) will delete everything, essentially resetting the
font to a non-SMuFL state. Individual attribute and anchor names may
otherwise be specified in both parameters as a single string or tuple
of any of the following values:

    Attributes
        - `designSize`
        - `sizeRange`
        - `engravingDefaults`
        - `name`
        - `classes`
        - `description`
        - `spaces`

    Anchors
        - `splitStemUpSE`
        - `splitStemUpSW`
        - `splitStemDownNE`
        - `splitStemDownNW`
        - `stemUpSE`
        - `stemDownNW`
        - `stemUpNW`
        - `stemDownSW`
        - `nominalWidth`
        - `numeralTop`
        - `numeralBottom`
        - `cutOutNE`
        - `cutOutSE`
        - `cutOutSW`
        - `cutOutNW`
        - `graceNoteSlashSW`
        - `graceNoteSlashNE`
        - `graceNoteSlashNW`
        - `graceNoteSlashSE`
        - `repeatOffset`
        - `noteheadOrigin`
        - `opticalCenter`

.. Warning:: This script will permanently delete data. Remember to
   always back up your file before running.

This script requires SMufoLib to be installed within its executive
environment. It may also be imported as a module and contains the
following public funcitons:

    - :func:`cleanFont` - The scripts program function.
    - :func:`main` - Command line entry point.

"""

from collections.abc import Iterable
import argparse
from pathlib import Path

from tqdm import tqdm

from smufolib import Font
from smufolib.objects.smufl import ANCHOR_NAMES, FONT_ATTRIBUTES, GLYPH_ATTRIBUTES
from smufolib import cli
from smufolib.utils import error, stdUtils
from smufolib.utils.scriptUtils import normalizeFont as _normalizeFont


# Parameter defaults
EXCLUDE = None
VERBOSE = False

# pylint: disable=C0103


def cleanFont(
    font: Font | Path | str,
    include: Iterable,
    exclude: Iterable | None = EXCLUDE,
    verbose: bool = VERBOSE,
):
    """Delete Smufl-specific attribute values.

    :param font: Object or path to
        target :class:`~smufolib.objects.font.Font`.
    :param include: items to be deleted. May be ``'*'`` (all), an
        individual attribute or anchor name as a :class:`str`
        or :class:`tuple` of several.
    :param exclude: Items to be preserved if ``ìnclude='*'``. Defaults
        to :obj:`None`.
    :param verbose: Make output verbose. Defaults to :obj:`False`.
    :raises TypeError: If any parameter value is not the expected type.
    :raises ValueError: If any item in `include` or `exclude` is
        invalid.

    """
    print("Starting...")

    font = _normalizeFont(font)
    itemsToClean = _buildItemsDict(include, exclude)

    # Clean font attributes
    stdUtils.verbosePrint("\nCleaning attributes for font:", verbose)
    for attr in itemsToClean["fontAttributes"]:
        if attr == "engravingDefaults":
            font.smufl.engravingDefaults.clear()
        elif attr == "spaces":
            setattr(font.smufl, attr, False)
        else:
            setattr(font.smufl, attr, None)
        stdUtils.verbosePrint(f"\t'{attr}'", verbose)

    for glyph in font if verbose else tqdm(font):
        # Clean glyph attributes
        glyphAttributesCleaned = False
        for attr in itemsToClean["glyphAttributes"]:
            if not getattr(glyph.smufl, attr):
                continue

            if not glyphAttributesCleaned:
                stdUtils.verbosePrint(
                    f"\nCleaning attributes from glyph '{glyph.name}':", verbose
                )
                glyphAttributesCleaned = True

            setattr(glyph.smufl, attr, None)
            stdUtils.verbosePrint(f"\t'{attr}'", verbose)

        # Clean Anchors
        anchorsCleaned = False
        for anchor in glyph.anchors:
            if anchor.name not in itemsToClean["anchors"]:
                continue

            if not anchorsCleaned:
                stdUtils.verbosePrint(
                    f"\nCleaning anchors from glyph '{glyph.name}':", verbose
                )
                anchorsCleaned = True

            glyph.removeAnchor(anchor.index)
            stdUtils.verbosePrint(f"\t'{anchor.name}'", verbose)

    stdUtils.verbosePrint("\nSaving font...", verbose)
    font.save()

    print("\nDone.")


def main() -> None:
    """Command line entry point."""
    args = _parseArgs()
    cleanFont(args.font, args.include, exclude=args.exclude, verbose=args.verbose)


def _buildItemsDict(include, exclude):
    # Build dict of attribute and anchor items to remove.

    itemsToClean = {"fontAttributes": [], "glyphAttributes": [], "anchors": []}

    allItems = FONT_ATTRIBUTES | GLYPH_ATTRIBUTES | ANCHOR_NAMES
    exclude = () if exclude is None else exclude
    exclude = (exclude,) if isinstance(exclude, str) else exclude

    for item in exclude:
        error.suggestValue(item, allItems, "exclude")

    if include in ("*", ["*"]):
        include = allItems
    elif isinstance(include, str):
        include = (include,)

    for item in include:
        if item in exclude:
            continue

        if item in FONT_ATTRIBUTES:
            itemsToClean["fontAttributes"].append(item)
        elif item in GLYPH_ATTRIBUTES:
            itemsToClean["glyphAttributes"].append(item)
        elif item in ANCHOR_NAMES:
            itemsToClean["anchors"].append(item)
        else:
            error.suggestValue(item, allItems, "include")

    return itemsToClean


def _parseArgs() -> argparse.Namespace:
    # Parse command line arguments and options.
    parser = cli.commonParser(
        "font",
        "include",
        description=stdUtils.getSummary(cleanFont.__doc__),
        exclude=EXCLUDE,
        verbose=VERBOSE,
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



# pylint: disable=C0103, C0114, R0904, W0212, W0221
from __future__ import annotations
from typing import TYPE_CHECKING, Any
import re

from fontParts.base.base import BaseObject
from smufolib.objects.range import Range
from smufolib.objects.engravingDefaults import EngravingDefaults
from smufolib.utils import converters, error, normalizers

if TYPE_CHECKING:  # pragma: no cover
    from smufolib.objects.layer import Layer
    from smufolib.objects.font import Font
    from smufolib.objects.glyph import Glyph

#: Names of glyph anchors specified by the SMuFL standard.
ANCHOR_NAMES: set[str] = {
    "splitStemUpSE",
    "splitStemUpSW",
    "splitStemDownNE",
    "splitStemDownNW",
    "stemUpSE",
    "stemDownNW",
    "stemUpNW",
    "stemDownSW",
    "nominalWidth",
    "numeralTop",
    "numeralBottom",
    "cutOutNE",
    "cutOutSE",
    "cutOutSW",
    "cutOutNW",
    "graceNoteSlashSW",
    "graceNoteSlashNE",
    "graceNoteSlashNW",
    "graceNoteSlashSE",
    "repeatOffset",
    "noteheadOrigin",
    "opticalCenter",
}


#: Names of font-specific attributes of the :class:`Smufl` class.
FONT_ATTRIBUTES: set[str] = {"designSize", "engravingDefaults", "sizeRange", "spaces"}

#: Names of glyph-specific attributes of the :class:`Smufl` class.
GLYPH_ATTRIBUTES: set[str] = {"classes", "description", "name"}


class Smufl(BaseObject):
    """Provide SMuFL-related metadata and functionality.

    This class may be called from either :class:`.Font`
    or :class:`.Glyph`. Font-specific attributes with unique names may
    be accessed from both, due to the consistent access to parent
    classes in FontParts.

    .. _about glyph naming:
    .. note:: Attributes with the purpose of identifying ligatures
       (:attr:`isLigature`), stylistic alternates (:attr:`isSalt`) and
       stylistic sets (:attr:`isSet`) depend on strict adherence to the
       descriptive naming schemes stipulated in the `Adobe Glyph List\
       Specification <https://github.com/adobe-type-tools/agl-\
       specification#readme>`_ and followed by the SMuFL standard
       (see `Section 6 <https://github.com/adobe-type-tools/agl-\
       specification#6-assigning-glyph-names-in-new-fonts>`_ for more
       information).

    .. tip:: To avoid having to set all the glyph identification
       attributes manually, it is advisable to run the
       script :mod:`.importID` prior to using this class with an
       excisting font for the first time.

    :param font: Parent :class:`.Font` object.
    :param glyph: Parent :class:`.Glyph` object.

    While this object is normally created as part of a :class:`.Font`,
    an orphan :class:`Smufl` object can be created like this::

        >>> smufl = Smufl()

    """

    def _init(self, font: Font | None = None, glyph: Glyph | None = None) -> None:
        self._font = font
        self._glyph = glyph
        self._layer = None

    def _reprContents(self) -> list[str]:
        contents = []
        if self._glyph is not None:
            contents.append("in glyph")
            contents += self.glyph._reprContents()  # type: ignore
        if self.font is not None:
            contents.append("in font")
            contents += self.font._reprContents()
        return contents

    def naked(self):
        # BaseObject override for __eq__ and __hash__
        return self

    # -------
    # Parents
    # -------

    @property
    def font(self) -> Font | None:
        """Parent :class:`.Font` object.

        Example::

            >>> glyph.smufl.font
            <Font 'MyFont' path='path/to/my/font.ufo' at 4405856720>

        """
        if self._font is not None:
            return self._font
        if self._glyph is not None:
            return self._glyph.font
        return None

    @font.setter
    def font(self, value: Font) -> None:
        if self._font is not None and self._font != value:
            raise AssertionError(
                "Font for Smufl object is already set and is not same as value."
            )
        if self._glyph is not None:
            raise AssertionError("Glyph for Smufl object is already set.")
        self._font = normalizers.normalizeFont(value)

    @property
    def glyph(self) -> Glyph | None:
        """Parent :class:`.Glyph` object.

        Example::

            >>> glyph.smufl.glyph
            <Glyph 'uniE587' ('public.default') at 4536458160>

        """
        return self._glyph

    @glyph.setter
    def glyph(self, value: Glyph) -> None:
        if self._font is not None:
            raise AssertionError("Font for Smufl object is already set.")
        if self._glyph is not None and self._glyph != value:
            raise AssertionError(
                "Glyph for Smufl object is already set and is not same as value."
            )
        self._glyph = normalizers.normalizeGlyph(value)

    @property
    def layer(self) -> Layer | None:
        """Parent :class:`.Layer` object.

        This property is read-only.

        Example::

            >>> glyph.smufl.layer
            <Layer 'public.default' at 4568631952>

        """
        if self._glyph is None:
            return None
        return self._glyph.layer

    # -------------
    # Font Metadata
    # -------------

    @property
    def designSize(self) -> int | None:
        """Optimum point size in integral decipoints.

        Example::

            >>> font.smufl.designSize
            240

        """
        if self.font is None:
            return None
        return self.font.lib.naked().get("com.smufolib.designSize", None)

    @designSize.setter
    def designSize(self, value: int | None) -> None:
        self._updateFontLib(
            "com.smufolib.designSize", normalizers.normalizeDesignSize(value)
        )

    @property
    def engravingDefaults(self) -> EngravingDefaults:
        """Font's :class:`.EngravingDefaults` object.

        Example::

            >>> font.smufl.engravingDefaults
            <EngravingDefaults at 4540156752>

        """
        return EngravingDefaults(self)

    @engravingDefaults.setter
    def engravingDefaults(self, value: EngravingDefaults) -> None:
        self.engravingDefaults.update(normalizers.normalizeEngravingDefaults(value))

    @property
    def sizeRange(self) -> tuple[int, int] | None:
        """Optimum size range in integral decipoints.

        Example::

            >>> font.smufl.designSize
            (180, 260)

        """
        if self.font is None:
            return None
        return self.font.lib.naked().get("com.smufolib.sizeRange", None)

    @sizeRange.setter
    def sizeRange(self, value: tuple[int, int] | None) -> None:
        self._updateFontLib(
            "com.smufolib.sizeRange", normalizers.normalizeSizeRange(value)
        )

    def _updateFontLib(self, key: str, value: Any) -> None:
        # Common font metadata setter.
        if self.font is not None:
            if value is None:
                self.font.lib.naked().pop(key)
            else:
                self.font.lib.naked()[key] = value

    # --------------
    # Glyph metadata
    # --------------

    @property
    def alternates(self) -> tuple[dict[str, str], ...] | None:
        """Metadata of glyph alternates.

        This property is read-only.

        Example::

            >>> glyph = font['uniE050'] # gClef
            >>> glyph.smufl.alternates
            ({'codepoint': 'U+F472', 'name': 'gClefSmall'},)

        """
        if self.glyph is None or self.font is None:
            return None
        # find alt names among string of glyph names
        string = " ".join(sorted(self.font.keys()))
        pattern = rf"\b{self.glyph.name}\.(?:s?alt|ss)[0-9]{{2}}\b"
        results = re.findall(pattern, string)
        alternates = []
        for name in results:
            glyph = self.font[name]
            alternates.append(
                {"codepoint": glyph.smufl.codepoint, "name": glyph.smufl.name}
            )
        return tuple(alternates)

    @property
    def anchors(self) -> dict[str, tuple[int | float, int | float]] | None:
        """SMuFL-specific glyph anchors as Cartesian coordinates.

        This property is read-only. Use
        the :attr:`fontParts.base.BaseGlyph.anchors` attribute to set
        glyph anchors.

        Example::

            >>> glyph = font['uniE0A3'] # noteheadHalf
            >>> glyph.smufl.anchors
            {'cutOutNW': (0.204, 0.296), 'cutOutSE':
            (0.98, -0.3), 'splitStemDownNE': (0.956, -0.3), 'splitStemDownNW':
            (0.128, -0.428), 'splitStemUpSE': (1.108, 0.372), 'splitStemUpSW':
            (0.328, 0.38), 'stemDownNW': (0.0, -0.168), 'stemUpSE': (1.18, 0.168)}

        """
        if self.glyph is None:
            return None

        anchors = {}
        for a in self.glyph.naked().anchors:
            if a.name in ANCHOR_NAMES:
                x = self.toSpaces(a.x) if self.spaces else a.x
                y = self.toSpaces(a.y) if self.spaces else a.y

                if x is None or y is None:
                    return None

                anchors[a.name] = (x, y)

        return anchors

    @property
    def bBox(self) -> dict[str, tuple[int | float, int | float]] | None:
        """Glyph bounding box as Cartesian coordinates.

        This property is read-only.

        Example::

            >>> glyph.smufl.bBox
            {'bBoxSW': (0.0, -0.5), 'bBoxNE': (1.18, 0.5)}

        """
        if self.glyph is None or not self.glyph.bounds:
            return None
        xMin, yMin, xMax, yMax = self.glyph.bounds
        if self.spaces:
            xMin, yMin, xMax, yMax = [self.toSpaces(b) for b in self.glyph.bounds]
        return {"bBoxSW": (xMin, yMin), "bBoxNE": (xMax, yMax)}

    @property
    def codepoint(self) -> str | None:
        """Unicode codepoint as formatted string.

        Example::

            >>> glyph = font['uniE0A3']
            >>> glyph.smufl.codepoint
            U+E0A3

        """
        if self.glyph is None or not self.glyph.unicode:
            return None
        return converters.toUniHex(self.glyph.unicode)

    @codepoint.setter
    def codepoint(self, value: str | None) -> None:
        if self.glyph is not None:
            if value is None:
                self.glyph.unicode = None
            else:
                self.glyph.unicode = converters.toDecimal(value)

    @property
    def componentGlyphs(self) -> tuple[Glyph, ...] | None:
        """Ligature components by :class:`.Glyph` object.

        This property is read-only.

        Example::

            >>> glyph = font['uniE09E_uniE083_uniE09F_uniE084']
            >>> glyph.smufl.componentGlyphs
            (<Glyph 'uniE09E' ('public.default') at 4399803376>,
             <Glyph 'uniE083' ('public.default') at 4399803184>,
             <Glyph 'uniE09F' ('public.default') at 4399797952>,
             <Glyph 'uniE084' ('public.default') at 4399797760>)

        """
        if self.glyph is None or self.font is None:
            return None
        if not self.isLigature:
            return ()

        components = [
            self.font[n] for n in self.glyph.name.split("_") if n in self.font
        ]
        return tuple(components)

    @property
    def componentNames(self) -> tuple[str | None, ...] | None:
        """Ligature components by :attr:`name`.

        This property is read-only.

        Example::

            >>> glyph = font['uniE09E_uniE083_uniE09F_uniE084']
            >>> glyph.smufl.componentNames
            ('timeSigCombNumerator', 'timeSig3',
             'timeSigCombDenominator', 'timeSig4')

        """
        if self.glyph is None or self.font is None:
            return None
        if not self.componentGlyphs:
            return ()

        components = [g.smufl.name for g in self.componentGlyphs]
        return tuple(components)

    @property
    def range(self) -> Range:
        """Glyph's :class:`.Range` object.

        This property is read-only.

        Example::

            >>> glyph = font['uniE212'] # stemSwished
            >>> glyph.smufl.range
            <Range 'stems' ('U+E210–U+E21F') at 4348391632>

        """
        return Range(self)

    @property
    def advanceWidth(self) -> int | float | None:
        """Glyph advance width.

        This property is equivalent
        to :attr:`fontParts.base.BaseGlyph.width`.

        Example::

            >>> glyph.smufl.advanceWidth
            671

        """
        if self.glyph is None:
            return None
        if self.spaces:
            return self.toSpaces(self.glyph.width)
        return self.glyph.width

    @advanceWidth.setter
    def advanceWidth(self, value: int | float | None) -> None:
        if self.glyph is not None:
            if self.spaces and value is not None:
                self.glyph.width = self.toUnits(value)
            else:
                self.glyph.width = value

    # --------------
    # Identification
    # --------------

    # Font
    # ----

    # Font family name is acessible through font.smufl.name.

    @property
    def version(self) -> float | None:
        """SMuFL-specific font version number.

        Example::

            >>> font.smufl.version
            2.2

        """
        if self.font is None:
            return None
        try:
            return float(
                f"{self.font.info.naked().versionMajor}."
                f"{self.font.info.naked().versionMinor}"
            )
        except ValueError:
            return None

    @version.setter
    def version(self, value: float | None) -> None:
        if value is None:
            major, minor = None, None
        else:
            major, minor = [int(n) for n in str(value).split(".")]
        if self.font is not None:
            self.font.info.naked().versionMajor = major
            self.font.info.naked().versionMinor = minor

    # Glyph
    # -----
    @property
    def classes(self) -> tuple[str, ...] | None:
        """SMuFL-specific class memberships.

        Example::

            >>> glyph = font['uniE354'] # accSagittalSharp7v11kUp
            >>> glyph.smufl.classes
            ['accidentals', 'accidentalsSagittalAthenian', 'combiningStaffPositions']

        """
        if self.glyph is None:
            return None
        return tuple(self.glyph.lib.naked().get("com.smufolib.classes", ()))

    @classes.setter
    def classes(self, value: tuple[str, ...] | None) -> None:
        self._updateGlyphLib(
            "com.smufolib.classes", normalizers.normalizeClasses(value)
        )

    @property
    def description(self) -> str | None:
        """SMuFL-specific glyph description.

        Example::

            >>> glyph.smufl.description
            Combining swished stem

        """
        if self.glyph is None:
            return None
        return self.glyph.lib.naked().get("com.smufolib.description", None)

    @description.setter
    def description(self, value: str | None) -> None:
        self._updateGlyphLib(
            "com.smufolib.description", normalizers.normalizeDescription(value)
        )

    @property
    def name(self) -> str | None:
        """SMuFL-specific canonical font or glyph name.

        Examples::

            >>> font.smufl.name
            Bravura

        ::

            >>> glyph = font['uniE212']
            >>> glyph.smufl.name
            stemSwished

        """
        if self.font is None:
            return None
        if self.glyph is None:
            return self.font.info.naked().familyName
        return self.glyph.lib.naked().get("com.smufolib.name", None)

    @name.setter
    def name(self, value: str | None) -> None:
        # Update com.smufolib.names before ID property
        if self.font is None:
            return

        if self.glyph is None:
            self.font.info.naked().familyName = value
        else:
            self._updateNames(normalizers.normalizeSmuflName(value))
            self._updateGlyphLib(
                "com.smufolib.name", normalizers.normalizeSmuflName(value)
            )

    def _updateGlyphLib(self, key: str, value: Any) -> None:
        if self.glyph is not None:
            if not value:
                if key in self.glyph.lib.naked():
                    del self.glyph.lib.naked()[key]
            else:
                self.glyph.lib.naked()[key] = value

    def _clearNames(self) -> None:
        if self.font is not None:
            if self._names:
                self.font.lib.naked()["com.smufolib.names"].pop(self.name, None)
            if not self._names:
                self.font.lib.naked().pop("com.smufolib.names", None)

    def _addNames(self, value: Any) -> None:
        if self._names is None:
            self._names = {}
        if self.glyph is not None:
            self._names[value] = self.glyph.name

    def _updateNames(self, value: str | None) -> None:
        # Keep dynamic dict of glyph names in font.lib.
        if value is None:
            self._clearNames()
        else:
            self._addNames(value)

    @property
    def _names(self) -> dict[str, str] | None:
        # Dict of glyph names in font.lib.
        if self.font is None:
            return None
        return self.font.lib.naked().get("com.smufolib.names")

    @_names.setter
    def _names(self, value: dict[str, str] | None) -> None:
        self._updateFontLib("com.smufolib.names", value)

    # ----------
    # Validation
    # ----------

    @property
    def isLigature(self) -> bool:
        """Return :obj:`True` if glyph is ligature.

        This property is read-only.

        Example::

            >>> glyph = font['uniE09E_uniE083_uniE09F_uniE084']
            >>> g1.smufl.isLigature
            True
            >>> glyph = font['uniE083']
            >>> glyph.smufl.isLigature
            False

        """
        if (
            self.glyph is not None
            and self.glyph.name
            and self.glyph.name.count("uni") > 1
            and "_" in self.glyph.name
        ):
            return True
        return False

    @property
    def isMember(self) -> bool:
        """Return :obj:`True` if glyph is either `recommended or optional
        <https://w3c.github.io/smufl/latest/about/recommended-chars-\
        optional-glyphs.html>`_.

        This property is read-only.

        Example::

            >>> glyph = font['uniE050']
            >>> glyph.smufl.isMember
            True
            >>> glyph = font['spaces']
            >>> glyph.smufl.isMember
            False

        """
        if (
            self.glyph is not None
            and self.glyph.unicode
            and 0xE000 <= self.glyph.unicode <= 0xF8FF
        ):
            return True
        return False

    @property
    def isOptional(self) -> bool:
        """Return :obj:`True` if glyph is `optional <https://w3c.github.io\
        /smufl/latest/about/recommended-chars-optional-glyphs.html>`_.

        This property is read-only.

        Example::

            >>> glyph = font['uniF660']
            >>> g1.smufl.isOptional
            True
            >>> glyph = font['uniE083']
            >>> glyph.smufl.isOptional
            False

        """
        if (
            self.glyph is not None
            and self.glyph.unicode
            and 0xF400 <= self.glyph.unicode <= 0xF8FF
        ):
            return True
        return False

    @property
    def isRecommended(self) -> bool:
        """Return :obj:`True` if glyph is `recommended <https://w3c.github.io\
        /smufl/latest/about/recommended-chars-optional-glyphs.html>`_.

        This property is read-only.

        Example::

            >>> glyph = font['uniE083']
            >>> g1.smufl.isRecommended
            True
            >>> glyph = font['uniF660']
            >>> glyph.smufl.isRecommended
            False

        """
        if (
            self.glyph is not None
            and self.glyph.unicode
            and 0xE000 <= self.glyph.unicode <= 0xF3FF
        ):
            return True
        return False

    @property
    def isSalt(self) -> bool:
        """Return :obj:`True` if glyph is stylistic alternate.

        Glyph names with either ``'.alt'`` and ``'.salt'`` suffix are
        accepted. See :ref:`Note <about glyph naming>` about glyph
        naming.

        This property is read-only.

        Example::

            >>> glyph = font['uniE042.salt01']
            >>> glyph.smufl.isSalt
            True
            >>> glyph = font['uniE042']
            >>> glyph.smufl.isSalt
            False

        """
        if (
            self.glyph is not None
            and self.glyph.name
            and (
                self.glyph.name.endswith(".salt", 7, -2)
                or self.glyph.name.endswith(".alt", 7, -2)
            )
        ):
            return True
        return False

    @property
    def isSet(self) -> bool:
        """Return :obj:`True` if glyph is stylistic set member.

        See :ref:`Note <about glyph naming>` about glyph naming.

        This property is read-only.

        Example::

            >>> glyph = font['uniE050.ss01']
            >>> glyph.smufl.isSet
            True
            >>> glyph = font['uniE042.salt01']
            >>> glyph.smufl.isSet
            False

        """
        if (
            self.glyph is not None
            and self.glyph.name
            and self.glyph.name.endswith(".ss", 7, -2)
        ):
            return True
        return False

    # -----------------------------
    # Normalization and Measurement
    # -----------------------------

    def round(self) -> None:
        """Round font units to integers.

        Method applies to the following attributes:

        - :attr:`Smufl.engravingDefaults`
        - :attr:`Smufl.anchors`
        - :attr:`Smufl.advanceWidth`
        - :attr:`.BaseGlyph.width`
        - :attr:`.BaseGlyph.height`
        - :attr:`.BaseGlyph.contours`
        - :attr:`.BaseGlyph.components`
        - :attr:`.BaseGlyph.anchors`
        - :attr:`.BaseGlyph.guidelines`


        If :attr:`spaces` is :obj:`True`, values are left unchanged.

        Examples::

            >>> glyph.smufl.advanceWidth
            230.5
            >>> glyph.smufl.round()
            >>> glyph.smufl.advanceWidthdth
            231

        ::

            >>> glyph.smufl.spaces = True
            >>> glyph.smufl.advanceWidth
            0.922
            >>> glyph.smufl.round()
            >>> glyph.smufl.advanceWidth
            0.922

        """
        if self.spaces:
            return
        self.engravingDefaults.round()
        if self.glyph is not None:
            self.glyph.round()

    def toSpaces(self, value: int | float) -> float | None:
        """Convert font units to staff spaces based on font UPM size.

        The inverse of :meth:`toUnits`.

        :param value: Value to convert.

        Example::

            >>> f = Font('path/to/my/font.ufo')
            >>> font.info.unitsPerEm
            2048
            >>> font.smufl.toSpaces(512)
            1.0

        """
        if self.font is None:
            return None

        if not self.font.info.unitsPerEm:
            raise AttributeError(
                error.generateErrorMessage(
                    "missingDependencyError",
                    objectName="value",
                    dependency="font.info.unitsPerEm",
                )
            )

        return converters.convertMeasurement(
            measurement=value,
            targetUnit="spaces",
            unitsPerEm=self.font.info.unitsPerEm,
            rounded=False,
        )

    def toUnits(self, value: int | float, rounded=True) -> int | float | None:
        """Convert staff spaces to font units based on font UPM size.

        The inverse of :meth:`toSpaces`.
        Result is always rounded.

        :param value: Value to convert.
        :param rounded: Whether to round result to nearest integer.

        Example::

            >>> f = Font('path/to/my/font.ufo')
            >>> font.info.unitsPerEm
            2048
            >>> font.smufl.toSpaces(2)
            1024

        """
        if self.font is None:
            return None

        if not self.font.info.unitsPerEm:
            raise AttributeError(
                error.generateErrorMessage(
                    "missingDependencyError",
                    objectName="value",
                    dependency="font.info.unitsPerEm",
                )
            )

        return converters.convertMeasurement(
            measurement=value,
            targetUnit="units",
            unitsPerEm=self.font.info.unitsPerEm,
            rounded=rounded,
        )

    @property
    def spaces(self) -> bool:
        """Set state of measurement to staff spaces.

        Example::

            >>> glyph.smufl.advanceWidth
            230.5
            >>> glyph.smufl.spaces = True
            >>> glyph.smufl.advanceWidth
            0.922

        """
        if self.font is None:
            return False
        return self.font.lib.naked().get("com.smufolib.spaces", False)

    @spaces.setter
    def spaces(self, value):
        if self.font is not None:
            if not self.font.info.unitsPerEm:
                raise AttributeError(
                    error.generateErrorMessage(
                        "missingDependencyError",
                        objectName="spaces",
                        dependency="font.info.unitsPerEm",
                    )
                )
            value = normalizers.normalizeBoolean(value)
            if value:
                self.font.lib.naked()["com.smufolib.spaces"] = True
            else:
                self.font.lib.naked().pop("com.smufolib.spaces", False)

    # -----
    # Other
    # -----

    @property
    def base(self) -> Glyph | None:
        """:class:`.Glyph` object of alternate's base glyph.

        This property is read-only.

        Example::

            >>> glyph = font['uniE050.ss01']
            >>> glyph.smufl.base
            <Glyph 'uniE050' ('public.default') at 4373577008>

        """
        baseName = self._getBasename()
        if self.font and baseName:
            return self.font[baseName]
        return None

    def _getBasename(self) -> str | None:
        # Get name of base glyph.
        if self.glyph is None or self.glyph.name is None:
            return None
        basename = self.glyph.name[:7]
        return basename if self.font and basename in self.font else None

    def findGlyph(self, name: str) -> Glyph | None:
        """Find :class:`.Glyph` object from :attr:`name`.

        :param name: SMuFL-specific canonical glyph name.

        Example::

            >>> font.smufl.findGlyph('barlineSingle')
            <Glyph 'uniE030' ('public.default') at 4393557200>

        """
        if name is None:
            return None

        normalizedName = normalizers.normalizeSmuflName(name)

        if (
            self.font is None
            or self._names is None
            or normalizedName not in self._names
        ):
            return None

        return self.font[self._names[normalizedName]]

    @property
    def suffix(self) -> str | None:
        """Return suffix of alternates.

        This property is read-only.

        Example::

            >>> glyph = font['uniE050.ss01']
            >>> glyph.smufl.suffix
            ss01

        """
        if self.glyph is not None and (self.isSalt or self.isSet):
            return self.glyph.name.split(".")[1]
        return None

    # ------------------------
    # Override from BaseObject
    # ------------------------

    def raiseNotImplementedError(self):
        """This exception needs to be raised frequently by
        the base classes. So, it's here for convenience.

        """
        raise NotImplementedError(  # pragma: no cover
            error.generateErrorMessage(
                "notImplementedError", objectName=self.__class__.__name__
            )
        )



"""Script utility functions.

This module provides a set of utility functions specifically intended for use in
scripts and applications. Particualarly, it provides a set of noramlizers to ensure the
integrity of objects frequently passed by the user in an application setting.

"""

from pathlib import Path
from typing import Any

from smufolib.objects.font import Font
from smufolib.request import Request
from smufolib.utils import error, normalizers


JsonDict = dict[str, Any]
ColorValue = int | float
ColorTuple = tuple[ColorValue, ColorValue, ColorValue, ColorValue]
ColorDict = dict[str, ColorTuple]


def normalizeColor(color: ColorTuple | None, mark: bool) -> ColorTuple | None:
    """Ensure that `color` is valid based on the `mark` setting."

        If the input `color` is :obj:`None` and `mark` is :obj:`True`, a :class:`TypeError`
        is raised. Otherwise, the color is normalized
    each dictionary value

        :param color: The color value to normalize.
        :param mark: Mark objects with the specified `color`.
        :raises TypeError: If `color` is not the expected type.

    """
    # Normalize `color` value.
    if color is None:
        if mark:
            raise TypeError(
                error.generateTypeError(
                    value=color,
                    validTypes=tuple,
                    objectName="color",
                    dependencyInfo="'mark' is True",
                )
            )
        return None
    return normalizers.normalizeColor(color)


def normalizeColorDict(colorDict: ColorDict | None, mark: bool) -> ColorDict | None:
    """Ensure that `ColorDict` is valid based on the `mark` setting."

    If the input `ColorDict` is :obj:`None` and `mark` is :obj:`True`,
    a :class:`TypeError` is raised. Otherwise, each value is normalized
    using :func:`.normalizeColor`.


    :param colorDict: The color dictionary to normalize.
    :param mark: Mark objects with the specified colors in `colorDict`.
    :raises TypeError: If `colorDict` is not the expected type.

    """
    if colorDict is None:
        if mark:
            raise TypeError(
                error.generateTypeError(
                    value=colorDict,
                    validTypes=ColorDict,
                    objectName="colors",
                    dependencyInfo="'mark' is True",
                )
            )
        return None

    error.validateType(colorDict, dict, "colors")
    for value in colorDict.values():
        normalizers.normalizeColor(value)

    return colorDict


def normalizeFont(font: Font | Path | str) -> Font:
    """Ensure that `font` represents a font object.

    If `font` is a :class:`str` or :class:`~pathlib.Path`, it is treated as a path and
    passed to the :class:`.Font` constructor.

    :param font: The font to normalize.
    :raises TypeError: If `font` is not the expected type.

    """
    # Convert font path to object if necessary.
    error.validateType(font, (Font, Path, str), "font")
    if isinstance(font, Font):
        return font
    return Font(font)


def normalizeJsonDict(jsonDict: JsonDict | None) -> JsonDict:
    """Ensure that `jsonDict` is not :obj:`None`.

    :param jsonDict: The JSON dictionary to validate.
    :raises TypeError: If `jsonDict` is :obj:`None`.

    """
    # Ensure `jsonDict` is not None.
    if jsonDict is None:
        raise TypeError(error.generateTypeError(jsonDict, JsonDict, "JSON file"))
    return jsonDict


def normalizeRequest(request: Request | Path | str) -> Request:
    """Ensure that `request` represents a request object.

    If `request` is a :class:`str` or :class:`~pathlib.Path`, it is treated as a path
    and passed to the :class:`.Request` constructor.

    :param request: The request to normalize.
    :raises TypeError: If `request` is not the expected type.

    """
    # Convert request path to object if necessary.
    error.validateType(request, (Request, Path, str), "request")
    if isinstance(request, Request):
        return request
    return Request(request)


def normalizeTargetPath(targetPath: str | Path) -> str | Path:
    # Ensure targetPath exists.
    if not Path(targetPath).exists():
        raise FileNotFoundError(
            error.generateErrorMessage("fileNotFound", objectName="targetPath")
        )
    return targetPath
